#!/usr/bin/env ruby

#
# $Id$
#
# Pref game model
#

#
# Player as such
#
module Xmlizable

  TAB_SIZE = 2
  NAMESPACE = 'http://prefcalc.garage.maemo.org'
  NAMESPACE_LIST = "pref:#{NAMESPACE}"

  def _xmlize(file, indent, str)
    file.puts((" " * TAB_SIZE * indent) + str)
  end

  def load_xml(document)
  end

end

class PlayerAccount
  include Xmlizable

  attr_reader :name

  def initialize(name)
    @name = name
  end

  def xmlize(file, indent)
    _xmlize(file, indent, "<pref:player name=\"#{@name}\"/>")
  end

  def self.load_player_names(document)
    player_names = Array.new
    document.find('/pref:game/pref:players/pref:player', NAMESPACE_LIST).each do |node|
      player_names.push(node.attributes['name'])
    end
    return player_names
  end
end

#
# All game information
#
class Game
  include Xmlizable

  attr_reader :accounts, :model, :variant, :rounds, :whist_price
  attr_accessor :pass_out_counter

  def initialize(variant, player_names, whist_price)
    @variant = variant
    @accounts = Array.new()
    player_names.each do |p| 
      @accounts.push(PlayerAccount.new(p))
    end
    case player_names.length
      when 3
        @model = GameModel::GM3.new
      when 4
        @model = GameModel::GM4.new
    end
    @variant.reset_pass_out(self)

    @listeners = Array.new()
    @rounds = Array.new()
    @dealer_idx = 0
    @whist_price = whist_price
  end

  def add_listener(l)
    @listeners.push(l)
  end

  def inspect
    s = "Total players: #{@accounts.length}\n"
    @accounts.each_index do |idx| 
      s += "  Player #{idx}:\n" + @accounts[idx].inspect
    end
    return s
  end

  # Array of all player names
  def names
    @accounts.map { |acc| acc.name }
  end

  # Total number of players in the game
  def num_players
    @accounts.length
  end

  # Find the player by name
  # Ugly implementation
  def idx_by_name(name)
    @accounts.each_index do |idx| 
      if @accounts[idx].name == name
        return idx
      end
    end
    return nil
  end

  def append(hand_results)
    @rounds.push(hand_results)

    @dealer_idx = (@dealer_idx + 1) % @accounts.length

    # Notify interested objects
    @listeners.each { |l| l.game_updated(hand_results) }
  end

  # Get the list of names of whisters, in proper order
  def get_whister_names(dealer_name, bid_winner_name)
    dnames = names
    dnames.delete_if { |name| name == dealer_name } if not @model.dealer_plays?
    dnames = dnames * 2
    dnames = dnames.slice(dnames.index(bid_winner_name) + 1, 2)
    dnames
  end

  # Get the list of ids of whisters, in proper order
  def get_whister_idxs(dealer_idx, bid_winner_idx)
    idxs = (0..@accounts.length-1).to_a
    idxs.delete_if { |idx| idx == dealer_idx } if not @model.dealer_plays?
    idxs = idxs * 2
    idxs = idxs.slice(idxs.index(bid_winner_idx) + 1, 2)
    idxs
  end

  def get_pass_out_names(dealer_name)
    dnames = names * 2
    dnames.delete_if { |name| name == dealer_name } if not @model.dealer_plays?
    idx = dnames.index(dealer_name)
    dnames = dnames.slice(idx.nil? ? 0 : idx + 1, 3)
    dnames
  end

  def get_pass_out_idxs(dealer_idx)
    idxs = (0..@accounts.length-1).to_a
    first_idx = idxs.index(dealer_idx)
    if @model.dealer_plays?
      first_idx += 1
    else
      idxs.delete_if { |idx| idx == dealer_idx } 
    end
    idxs = idxs * 2
    idxs = idxs.slice(first_idx, 3)
    idxs
  end

  def dealer
    @accounts[@dealer_idx]
  end

  def xmlize(file, indent)
    _xmlize(file, indent, "<pref:game xmlns:pref=\"#{NAMESPACE}\" " +
                                     "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" " +
                                     "xsi:noNamespaceSchemaLocation=\"prefcalc.xsd\">")
    _xmlize(file, indent + 1, "<pref:players>")
    @accounts.each { |a| a.xmlize(file, indent + 2) }
    _xmlize(file, indent + 1, "</pref:players>")
    @variant.xmlize(file, indent + 1)
    _xmlize(file, indent + 1, "<pref:whistPrice>#{@whist_price}</pref:whistPrice>")
    _xmlize(file, indent + 1, "<pref:rounds>")
    @rounds.each do |r|
      _xmlize(file, indent + 2, "<pref:round>")
      r.xmlize(file, indent + 3)
      _xmlize(file, indent + 2, "</pref:round>")
    end
    _xmlize(file, indent + 2, "</pref:rounds>")
    _xmlize(file, indent, "</pref:game>")
  end

  def self.load_xml(document)
    player_names = PlayerAccount.load_player_names(document)
    variant = GameVariant::load_xml(document)
    return Game.new(variant, player_names, 1)
  end

  def load_rounds(document)
    result_types = GameModel::HandResult::types
    document.find('/pref:game/pref:rounds/pref:round', NAMESPACE_LIST).each do |round|
      result_types.each do |clazz| 
        found_tags = round.find(clazz.tag, NAMESPACE_LIST)
        if (found_tags.size > 0)
          results = clazz.load_xml(self, round)
          if (results)
            append(results)
          end
          break
        end
      end
    end
  end
end

#
# Game model
#
class GameModel
  include GetText

  # What kind of game
  class Type
    POSITIVE = 0
    MIZERE = 1
    PASS_OUT = 2
  end

  TYPE_NAME = ["Positive", "Mizere", "Pass Out"]

  # Whether dealer is allowed to play
  def dealer_plays?
    nil
  end

  # Check the rules whether combination of tricks, defence flags etc is valid
  def validate_positive_game(bid, whister1_tricks, whister2_tricks,
                             defending_whister1, defending_whister2)
    if (!(defending_whister1 || defending_whister2) &&
         (whister1_tricks > 0 || whister2_tricks > 0))
      return Hash[HandResult::VALIDATION_RESULT=>false,
                  HandResult::VALIDATION_MESSAGE=>_('Some tricks are given to whisters without defence')]
    end
    if (whister1_tricks + whister2_tricks > 10)
      return Hash[HandResult::VALIDATION_RESULT=>false,
                  HandResult::VALIDATION_MESSAGE=>_('Total number of whists taken cannot exceed 10')]
    end
    return true
  end

  # Checks whether the tricks compose a full pass-out game
  def validate_pass_out(dealer_tricks, player1_tricks, player2_tricks, player3_tricks)
    result = ((dealer_tricks + player1_tricks + player2_tricks + player3_tricks) == 10)
    return result ? true :
                    Hash[HandResult::VALIDATION_RESULT=>false,
                         HandResult::VALIDATION_MESSAGE=>_('Total number of whists taken must be equal 10')]
  end

  #
  # Data structure for keeping the game results
  #
  class HandResult

    include Xmlizable

    VALIDATION_RESULT = 'if_valid'
    VALIDATION_MESSAGE = 'validation_message'

    attr_reader :game
    attr_accessor :type, :dealer, :mount_tricks_for_dealer, :tricks_on_bid_winner_for_dealer

    @@subclasses = Array.new()

    def self.inherited(clazz)
      @@subclasses.push(clazz)
    end

    def self.types
      return @@subclasses
    end

    def initialize(game, t, d, mtfd, tobwfd)
      @game = game
      @type = t
      @dealer = d
      @mount_tricks_for_dealer = mtfd
      @tricks_on_bid_winner_for_dealer = tobwfd
    end

    def _xmlize_std_attributes
      mtfd = @mount_tricks_for_dealer == 0 ? "" : " mountTricksForDealer=\"#{@mount_tricks_for_dealer}\""
      tobwfd = @tricks_on_bid_winner_for_dealer == 0 ? "" : " tricksOnBidWinnerForDealer=\"#{@tricks_on_bid_winner_for_dealer}\""
      return "dealer=\"#{@game.accounts[@dealer].name}\"#{mtfd}#{tobwfd}"
    end

    def self._load_std_attributes(game, round)
       mount_tricks_for_dealer = get_data(round, "@mountTricksForDealer").to_i
       tricks_on_bid_winner_for_dealer = get_data(round, "@tricksOnBidWinnerForDealer").to_i
       dealer = game.idx_by_name(get_data(round, "@dealer"))
       return [ dealer, mount_tricks_for_dealer, tricks_on_bid_winner_for_dealer ]
    end


    def self.get_data(round, relpath)
      node = round.find_first("#{tag}/#{relpath}")
      return node ? node.child.to_s : nil
    end

    def inspect
      mt = @mount_tricks_for_dealer == 0 ? "" : format("mount tricks %s", @mount_tricks_for_dealer)
      tobw = @tricks_on_bid_winner_for_dealer == 0 ? "" : format("tricks on bid winner %s", @tricks_on_bid_winner_for_dealer)
      if (mt == "")
        if (tobw == "")
          dt = ""
        else
          dt = format("(%s)", tobw)
        end
      else
        if (tobw == "")
          dt = format("(%s)", mt)
        else
          dt = format("(%s, %s)", mt, tobw)
        end
      end
      return format('Type: %s, dealer: "%s"%s',
                    GameModel::TYPE_NAME[@type],
                    @game.accounts[@dealer].name,
                    dt)
    end

    class Positive < HandResult
      attr_accessor :declared_game, :bid_winner
      attr_accessor :tricks_by_whister1, :tricks_by_whister2
      attr_accessor :whister1_defending, :whister2_defending
      attr_accessor :half_whist_to

      def self.tag
        'pref:positiveGame'
      end

      def initialize(game, d, mtfd, tobwfd, dg, bw, tbw1, tbw2, w1d, w2d, hw2)
        super(game, Type::POSITIVE, d, mtfd, tobwfd)
        @declared_game = dg
        @bid_winner = bw
        @tricks_by_whister1 = tbw1
        @tricks_by_whister2 = tbw2
        @whister1_defending = w1d
        @whister2_defending = w2d
        @half_whist_to = hw2
      end

      def inspect
        widxs = @game.get_whister_idxs(@dealer, @bid_winner)
        wn1 = @game.accounts[widxs[0]].name
        wn2 = @game.accounts[widxs[1]].name
        hw2 = @half_whist_to.nil? ? 'nobody' : @game.accounts[@half_whist_to].name
        return super + format(', declared game: %s, bid winner: "%s", ' +
                               'tricks by "%s": %s%s, tricks by "%s": %s%s, ' +
                               'half whist to %s',
                              @declared_game, 
                              @game.accounts[@bid_winner].name,
                              wn1, @tricks_by_whister1, @whister1_defending ? " (defending)" : "",
                              wn2, @tricks_by_whister2, @whister2_defending ? " (defending)" : "",
                              hw2)
      end

      def xmlize(file, indent)
         hwattr = @half_whist_to.nil? ? "" : "halfWhistTo=\"#{@game.accounts[@half_whist_to].name}\""
         _xmlize(file, indent, "<#{Positive::tag} #{_xmlize_std_attributes} declaredGame=\"#{@declared_game}\" bidWinner=\"#{@game.accounts[@bid_winner].name}\" #{hwattr}>")
         _xmlize(file, indent + 1, "<pref:whister tricks=\"#{@tricks_by_whister1}\" defending=\"#{@whister1_defending}\"/>")
         _xmlize(file, indent + 1, "<pref:whister tricks=\"#{@tricks_by_whister2}\" defending=\"#{@whister2_defending}\"/>")
         _xmlize(file, indent, "</#{Positive::tag}>")
      end

      def self.load_xml(game, round)
         dealer, mount_tricks_for_dealer, tricks_on_bid_winner_for_dealer = _load_std_attributes(game, round)

         declared_game = get_data(round, "@declaredGame").to_i
         bid_winner = game.idx_by_name(get_data(round, "@bidWinner"))
         half_whist_to = game.idx_by_name(get_data(round, "@halfWhistTo"))
         tbw1 = get_data(round, "pref:whister[1]/@tricks").to_i
         w1d = "true".eql?(get_data(round, "pref:whister[1]/@defending"))
         tbw2 = get_data(round, "pref:whister[2]/@tricks").to_i
         w2d = "true".eql?(get_data(round, "pref:whister[2]/@defending"))
         return Positive.new(game, dealer, 
                             mount_tricks_for_dealer, tricks_on_bid_winner_for_dealer,
                             declared_game, bid_winner,
                             tbw1, tbw2, w1d, w2d, half_whist_to)
      end
    end

    class Mizere < HandResult
      attr_accessor :player
      attr_accessor :tricks_by_player

      def self.tag
        'pref:mizere'
      end

      def initialize(game, d, mtfd, tobwfd, p, tbp)
        super(game, Type::MIZERE, d, mtfd, tobwfd)
        @player = p
        @tricks_by_player = tbp
      end

      def inspect
        return super + format(', mizere player: "%s" (%s tricks)', 
                              @game.accounts[@player].name, @tricks_by_player)
      end

      def xmlize(file, indent)
         _xmlize(file, indent, "<#{Mizere::tag} #{_xmlize_std_attributes} player=\"#{@game.accounts[@player].name}\" tricksByPlayer=\"#{@tricks_by_player}\"/>")
      end

      def self.load_xml(game, round)
         dealer, mount_tricks_for_dealer, tricks_on_bid_winner_for_dealer = _load_std_attributes(game, round)

         player = game.idx_by_name(get_data(round, "@player"))
         tricks_by_player = get_data(round, "@tricks_by_player").to_i
         return Mizere.new(game, dealer, 
                           mount_tricks_for_dealer, tricks_on_bid_winner_for_dealer,
                           player, tricks_by_player)
      end
    end

    class PassOut < HandResult
      attr_accessor :tricks_by_player1
      attr_accessor :tricks_by_player2
      attr_accessor :tricks_by_player3

      def self.tag
        'pref:passOut'
      end

      def initialize(game, d, mtfd, tbp1, tbp2, tbp3)
        super(game, Type::PASS_OUT, d, mtfd, 0)
        @tricks_by_player1 = tbp1
        @tricks_by_player2 = tbp2
        @tricks_by_player3 = tbp3
      end

      def inspect
        idxs = @game.get_pass_out_idxs(@dealer)
        return super + format(', tricks by "%s": %s, tricks by "%s": %s, tricks by "%s": %s', 
                              @game.accounts[idxs[0]].name, @tricks_by_player1,
                              @game.accounts[idxs[1]].name, @tricks_by_player2,
                              @game.accounts[idxs[2]].name, @tricks_by_player3)
      end

      def xmlize(file, indent)
        idxs = @game.get_pass_out_idxs(@dealer)
         _xmlize(file, indent, "<#{PassOut::tag} #{_xmlize_std_attributes}>")
         _xmlize(file, indent + 1, "<pref:tricks player=\"#{@game.accounts[idxs[0]].name}\">#{@tricks_by_player1}</pref:tricks>")
         _xmlize(file, indent + 1, "<pref:tricks player=\"#{@game.accounts[idxs[1]].name}\">#{@tricks_by_player2}</pref:tricks>")
         _xmlize(file, indent + 1, "<pref:tricks player=\"#{@game.accounts[idxs[2]].name}\">#{@tricks_by_player3}</pref:tricks>")
         _xmlize(file, indent, "</#{PassOut::tag}>")
      end

      def self.load_xml(game, round)
         dealer, mount_tricks_for_dealer, tricks_on_bid_winner_for_dealer = _load_std_attributes(game, round)

         player = get_data(round, "@player")
         tbp1 = get_data(round, "pref:tricks[1]").to_i
         tbp2 = get_data(round, "pref:tricks[2]").to_i
         tbp3 = get_data(round, "pref:tricks[3]").to_i
         return PassOut.new(game, dealer, 
                            mount_tricks_for_dealer,
                            tbp1, tbp2, tbp3)
      end
    end

  end

  class GM3 < GameModel
    def dealer_plays?
      true
    end
  end

  class GM4 < GameModel
    def dealer_plays?
      false
    end
  end
end

#
# Variations of some rules
# Abstract class
#
class GameVariant
  include Xmlizable

  # How much one trick gives into whists
  attr_accessor :base_TP_on_game_for_whister

  # How much goes to mount for taken trick on pass_out (on minimal game)
  attr_accessor :base_TP_on_pass_out

  # How much goes to the pool for a game
  attr_accessor :base_game_price

  # How much goes to mount for untaken trick
  attr_accessor :base_UTP_on_game_for_player

  # How much goes to mount for untaken trick
  attr_accessor :base_UTP_on_game_for_whister

  # Initial 'game' for pass out
  attr_accessor :pass_out_counter_initial

  # If pass out 'game' should be incremented after each pass out
  attr_accessor :pass_out_counter_incremental

  # Maximal pass out 'game' (if counter is incrementable)
  attr_accessor :pass_out_counter_maximum

  # If the pass out counter should wrap (from max to initial)
  attr_accessor :pass_out_counter_wrap

  # If true, on failed game whists are divided equally
  # otherwise, only defending whister gets real whists (passer gets only "difference")
  attr_accessor :is_whist_gentlemanish

  # How much goes to mount for one trick on pass-out
  # (depending on the game state)
  def pass_out_TP(game)
    (game.pass_out_counter - 5) * @base_TP_on_pass_out
  end

  # Reset the game state (for pass-outs)
  def reset_pass_out(game)
    game.pass_out_counter = @pass_out_counter_initial
  end

  # Change the game state (for pass-outs)
  def another_pass_out(game)
    if @pass_out_counter_incremental
      if (game.pass_out_counter < @pass_out_counter_maximum)
        game.pass_out_counter = game.pass_out_counter + 1;
      else
        if (@pass_out_counter_wrap)
          reset_pass_out(game)
        end
      end
    end
  end

  def xmlize(file, indent)
    _xmlize(file, indent, "<pref:conventions>")
    _xmlize(file, indent + 1, "<pref:baseGamePrice>#{base_game_price}</pref:baseGamePrice>")
    _xmlize(file, indent + 1, "<pref:takenTrickPrices>")
    _xmlize(file, indent + 2, "<pref:baseTPOnPositiveGameForWhister>#{base_TP_on_game_for_whister}</pref:baseTPOnPositiveGameForWhister>")
    _xmlize(file, indent + 2, "<pref:baseTPOnPassOut>#{base_TP_on_pass_out}</pref:baseTPOnPassOut>")
    _xmlize(file, indent + 1, "</pref:takenTrickPrices>")
    _xmlize(file, indent + 1, "<pref:untakenTrickPrices>")
    _xmlize(file, indent + 2, "<pref:baseUTPOnPositiveGameForPlayer>#{base_UTP_on_game_for_player}</pref:baseUTPOnPositiveGameForPlayer>")
    _xmlize(file, indent + 2, "<pref:baseUTPOnPositiveGameForWhister>#{base_UTP_on_game_for_whister}</pref:baseUTPOnPositiveGameForWhister>")
    _xmlize(file, indent + 1, "</pref:untakenTrickPrices>")
    _xmlize(file, indent + 1, "<pref:passOutCounter initial=\"#{pass_out_counter_initial}\" incremental=\"#{pass_out_counter_incremental}\" maximum=\"#{pass_out_counter_maximum}\" wrap=\"#{pass_out_counter_wrap}\"/>")
    _xmlize(file, indent + 1, "<pref:isWhistGentlemanish>#{is_whist_gentlemanish}</pref:isWhistGentlemanish>")
    _xmlize(file, indent, "</pref:conventions>")
  end

  def self.get_data(document, relpath)
    base = '/pref:game/pref:conventions'
    return document.find_first(base + relpath, NAMESPACE_LIST).child.to_s
  end

  def self.load_xml(document)
    gv = GameVariant.new
    gv.base_game_price =              get_data(document, '/pref:baseGamePrice').to_i
    gv.base_TP_on_game_for_whister =  get_data(document, '/pref:takenTrickPrices/pref:baseTPOnPositiveGameForWhister').to_i
    gv.base_TP_on_pass_out =          get_data(document, '/pref:takenTrickPrices/pref:baseTPOnPassOut').to_i
    gv.base_UTP_on_game_for_player =  get_data(document, '/pref:untakenTrickPrices/pref:baseUTPOnPositiveGameForPlayer').to_i
    gv.base_UTP_on_game_for_whister = get_data(document, '/pref:untakenTrickPrices/pref:baseUTPOnPositiveGameForWhister').to_i
    gv.pass_out_counter_initial =     get_data(document, '/pref:passOutCounter/@initial').to_i
    gv.pass_out_counter_incremental = get_data(document, '/pref:passOutCounter/@incremental').eql?("true")
    gv.pass_out_counter_maximum =     get_data(document, '/pref:passOutCounter/@maximum').to_i
    gv.pass_out_counter_wrap =        get_data(document, '/pref:passOutCounter/@wrap').eql?("true")
    gv.is_whist_gentlemanish =        get_data(document, '/pref:isWhistGentlemanish').eql?("true")
    return gv
  end

end

