#
# $Id$
#
# GUI Actions
#

require 'gtk2'
require 'new_game_dialog'
require 'hand_results_dialog'
require 'model/pref_calc_model_variants'
require 'xml'

class PrefCalcAction
  include GetText

  def initialize
    @listeners = Array.new
  end

  def add_listener(listener)
    @listeners.push(listener)
  end

  def do
    raise 'Undefined action'
  end

  def sensitive=(is_sensitive)
    @listeners.each{ |l| l.sensitive = is_sensitive }
  end

  def add_button(btn)
    btn.signal_connect('clicked') { self.do }
    add_listener(btn);
  end
end 

$action_game_new = PrefCalcAction.new
$action_game_save = PrefCalcAction.new
$action_game_load = PrefCalcAction.new

$action_round_new = PrefCalcAction.new
$action_about = PrefCalcAction.new
$action_quit = PrefCalcAction.new

$action_page_game = PrefCalcAction.new
$action_page_totals = PrefCalcAction.new
$action_page_history = PrefCalcAction.new

def $action_game_new.do
  dialog = NewGameDialog.new
  dialog.widget.run do |response|
    case response
      when Gtk::Dialog::RESPONSE_OK
        case dialog.game_variant
          when NewGameDialog::LENINGRAD
            variant=Leningrad.new
          when NewGameDialog::ROSTOV
            variant=Rostov.new
          when NewGameDialog::SOCHI
            variant=Sochi.new
        end
        $main_app_window.game = Game.new(variant, dialog.player_names, dialog.whist_price)
        save_current_game
        $action_round_new.sensitive = true
        $action_game_save.sensitive = true
    end
    dialog.widget.destroy
  end
end

def $action_game_save.do
  dialog = Gtk::FileChooserDialog.new("Save File",
                                       $main_app_window,
                                       Gtk::FileChooser::ACTION_SAVE,
                                       nil,
                                       [Gtk::Stock::CANCEL, Gtk::Dialog::RESPONSE_CANCEL],
                                       [Gtk::Stock::SAVE, Gtk::Dialog::RESPONSE_ACCEPT])

  set_filters(dialog)
  dialog.current_folder  = GLib.home_dir

  if dialog.run == Gtk::Dialog::RESPONSE_ACCEPT
    if File.exists?(dialog.filename)
      qdialog = Gtk::MessageDialog.new($main_app_window, 
                                       Gtk::Dialog::DESTROY_WITH_PARENT,
                                       Gtk::MessageDialog::QUESTION,
                                       Gtk::MessageDialog::BUTTONS_YES_NO,
                                       "The file #{dialog.filename} already exists. Overwite?")
      qdialog.run do |response|
        case response
          when Gtk::Dialog::RESPONSE_YES
            save_game(dialog.filename)
        end
      end
      qdialog.destroy
    else
      save_game(dialog.filename)
    end
  end
  dialog.destroy
end

def $action_game_load.do
  dialog = Gtk::FileChooserDialog.new("Open File",
                                       $main_app_window,
                                       Gtk::FileChooser::ACTION_OPEN,
                                       nil,
                                       [Gtk::Stock::CANCEL, Gtk::Dialog::RESPONSE_CANCEL],
                                       [Gtk::Stock::OPEN, Gtk::Dialog::RESPONSE_ACCEPT])

  set_filters(dialog)
  dialog.current_folder  = GLib.home_dir

  if dialog.run == Gtk::Dialog::RESPONSE_ACCEPT
    schema_document = XML::Document.file(PrefUtil::xsd_filename)
    schema = XML::Schema.document(schema_document)
    begin
      instance = XML::Document.file(dialog.filename)
      validation_result = instance.validate_schema(schema)
      if !validation_result
        errDlg = Gtk::MessageDialog.new(dialog, 
                         Gtk::Dialog::DESTROY_WITH_PARENT,
                         Gtk::MessageDialog::ERROR,
                         Gtk::MessageDialog::BUTTONS_CLOSE,
                         "The document has invalid structure")
        errDlg.run
        errDlg.destroy
      else
        $main_app_window.game = Game::load_xml(instance)
        $main_app_window.game.load_rounds(instance)
        save_current_game
        $action_round_new.sensitive = true
        $action_game_save.sensitive = true
      end
    rescue Exception => e
      errDlg = Gtk::MessageDialog.new(dialog, 
                       Gtk::Dialog::DESTROY_WITH_PARENT,
                       Gtk::MessageDialog::ERROR,
                       Gtk::MessageDialog::BUTTONS_CLOSE,
                       "Error reading the document:\n#{e}")
      errDlg.run
      errDlg.destroy
    end
  end
  dialog.destroy
end

def $action_round_new.do
  dialog = HandResultsDialog.new($main_app_window.game)
  dialog.widget.run
  if ! dialog.results.nil?
    $main_app_window.game.append(dialog.results)
    save_current_game
  end
end
 
def $action_about.do
  about_box = Gtk::AboutDialog.new()
  about_box.name = $APP_NAME
  about_box.version = $APP_VERSION
  about_box.copyright = "Copyright \xc2\xa9 2007-2009 Sergey V. Udaltsov"
  about_box.authors = [ "Sergey V. Udaltsov <svu@gnome.org>" ]
  about_box.license = $GPL_TEXT
  about_box.wrap_license = true
  about_box.website = 'http://prefcalc.garage.maemo.org'
  about_box.translator_credits = _('TRANSLATOR')
  about_box.modal = true
  about_box.transient_for = $main_app_window
  about_box.signal_connect('response') { about_box.destroy }

  about_box.show
end

def $action_quit.do
  Gtk.main_quit
end

def $action_page_game.do
  $main_app_window.notebook.page = 0
  $action_page_game.sensitive = false
  $action_page_totals.sensitive = true
  $action_page_history.sensitive = true
end

def $action_page_totals.do
  $main_app_window.notebook.page = 1
  $action_page_game.sensitive = true
  $action_page_totals.sensitive = false
  $action_page_history.sensitive = true
end

def $action_page_history.do
  $main_app_window.notebook.page = 2
  $action_page_game.sensitive = true
  $action_page_totals.sensitive = true
  $action_page_history.sensitive = false
end

def save_current_game
  private_dir = "#{ENV['HOME']}/.prefcalc"
  current_file = "#{private_dir}/current.pc"
  backup_file = "#{private_dir}/backup.pc"

  File.rename(current_file, backup_file) if File.file?(current_file)
  Dir.mkdir(private_dir) unless File.directory?(private_dir)

  save_game(current_file)
end

def save_game(filename)
  f = File.new(filename, "w")
  $main_app_window.game.xmlize(f, 0)
  f.close
end

def set_filters(dialog)
  filter1 = Gtk::FileFilter.new
  filter1.name='Saved Games'
  filter1.add_pattern('*.pc')
  filter1.add_pattern('*.xml')
  filter2 = Gtk::FileFilter.new
  filter2.name='All files'
  filter2.add_pattern('*')

  dialog.add_filter(filter1)
  dialog.add_filter(filter2)
end

